#define VERSION "0.1"

#include <math.h>
#include <complex.h>
#include <fftw3.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include "navier-stokes.h"
#include "driving.h"

// usage message
int print_usage();

// read command line arguments
int read_args(int argc, const char* argv[], char** params, unsigned int* driving_force, unsigned int* command, unsigned int* nthreads);
int read_params(char* params, int* K1, int* K2, int* N1, int* N2, unsigned int* nsteps, double* nu, double* delta, double* L, unsigned int* print_freq);
int set_parameter(char* lhs, char* rhs, int* K1, int* K2, int* N1, int* N2, unsigned int* nsteps, double* nu, double* delta, double* L, unsigned int* print_freq, bool* setN1, bool* setN2);


#define COMMAND_UK 1
#define COMMAND_ENSTROPHY 2
#define COMMAND_QUIET 3
#define COMMAND_ENERGY 4

#define DRIVING_ZERO 1
#define DRIVING_TEST 2


int main (
  int argc,
  const char* argv[]
){
  char* params=NULL;
  int K1,K2;
  int N1,N2;
  unsigned int nsteps;
  double nu,delta,L;
  _Complex double (*g)(int,int);
  int ret;
  unsigned int driving,command;
  unsigned int print_freq;
  unsigned int nthreads=1;

  command=0;
  driving=0;

  // read command line arguments
  ret=read_args(argc, argv, &params, &driving, &command, &nthreads);
  if(ret<0){
    return(-1);
  }
  // read params
  ret=read_params(params, &K1, &K2, &N1, &N2, &nsteps, &nu, &delta, &L, &print_freq);
  if(ret<0){
    return(-1);
  }
  // set driving force
  switch(driving){
  case DRIVING_ZERO:
    g=g_zero;
    break;
  case DRIVING_TEST:
    g=g_test;
    break;

  default:
    g=g_zero;
    break;
  }

  // run command
  if (command==COMMAND_UK){
    uk(K1, K2, N1, N2, nsteps, nu, delta, L, g, print_freq, nthreads);
  }
  else if (command==COMMAND_ENERGY){
    energy(K1, K2, N1, N2, nsteps, nu, delta, L, g, print_freq, nthreads);
  }
  else if(command==COMMAND_ENSTROPHY){
    enstrophy(K1, K2, N1, N2, nsteps, nu, delta, L, g, print_freq, nthreads);
  }
  else if(command==COMMAND_QUIET){
    quiet(K1, K2, N1, N2, nsteps, nu, delta, L, g, nthreads);
  }
  else if(command==0){
    fprintf(stderr, "error: no command specified\n");
    print_usage();
  }

  return(0);
}

// usage message
int print_usage(){
  fprintf(stderr, "usage:\n       nstrophy [-p parameters] [-g driving_force] <command>\n\n");
  return(0);
}

// read command line arguments
#define CP_FLAG_PARAMS 1
#define CP_FLAG_DRIVING 2
#define CP_FLAG_NTHREADS 3
int read_args(
  int argc,
  const char* argv[],
  char** params,
  unsigned int* driving_force,
  unsigned int* command,
  unsigned int* nthreads
){
  int i;
  int ret;
  // pointers
  char* ptr;
  // flag that indicates what argument is being read
  int flag=0;

  // loop over arguments
  for(i=1;i<argc;i++){
    // flag
    if(argv[i][0]=='-'){
      for(ptr=((char*)argv[i])+1;*ptr!='\0';ptr++){
	switch(*ptr){
	case 'p':
	  flag=CP_FLAG_PARAMS;
	  break;
	case 'g':
	  flag=CP_FLAG_DRIVING;
	  break;
	case 't':
	  flag=CP_FLAG_NTHREADS;
	  break;
	default:
	  fprintf(stderr, "unrecognized option '-%c'\n", *ptr);
	  print_usage();
	  return(-1);
	  break;
	}
      }
    }
    // params
    else if(flag==CP_FLAG_PARAMS){
      *params=(char*)argv[i];
      flag=0;
    }
    // driving force
    else if(flag==CP_FLAG_DRIVING){
      if (strcmp(argv[i],"zero")==0){
	*driving_force=DRIVING_ZERO;
      }
      else if (strcmp(argv[i],"test")==0){
	*driving_force=DRIVING_TEST;
      }
      else{
	fprintf(stderr, "error: unrecognized driving force '%s'\n",argv[i]);
	return(-1);
      }
      flag=0;
    }
    // nthreads
    else if(flag==CP_FLAG_NTHREADS){
      ret=sscanf(argv[i],"%u",nthreads);
      if(ret!=1){
	fprintf(stderr, "error: '-t' should be followed by an unsigned integer\n       got '%s'\n",argv[i]);
	return(-1);
      }
      flag=0;
    }
    // computation to run
    else{
      if(strcmp(argv[i], "uk")==0){
	*command=COMMAND_UK;
      }
      else if(strcmp(argv[i], "enstrophy")==0){
	*command=COMMAND_ENSTROPHY;
      }
      else if(strcmp(argv[i], "quiet")==0){
	*command=COMMAND_QUIET;
      }
      else if(strcmp(argv[i], "energy")==0){
	*command=COMMAND_ENERGY;
      }
      else{
	fprintf(stderr, "error: unrecognized command: '%s'\n",argv[i]);
	return(-1);
      }
      flag=0;
    }
  }

  return(0);
}

// read parameters string
int read_params(
  char* params,
  int* K1,
  int* K2,
  int* N1,
  int* N2,
  unsigned int* nsteps,
  double* nu,
  double* delta,
  double* L,
  unsigned int* print_freq
){
  int ret;
  // pointer in params
  char* ptr;
  // buffer and associated pointer
  char *buffer_lhs, *lhs_ptr;
  char *buffer_rhs, *rhs_ptr;
  // whether N was set explicitly
  bool setN1=false;
  bool setN2=false;
  // whether lhs (false is rhs)
  bool lhs=true;

  // defaults
  *K1=16;
  *K2=*K1;
  //delta=2^-13
  *delta=0.0001220703125;
  //nu=2^-11
  *nu=0.00048828125;
  *L=2*M_PI;
  *nsteps=10000000;
  *print_freq=1000;

  if (params!=NULL){
    // init
    buffer_lhs=calloc(sizeof(char),strlen(params));
    lhs_ptr=buffer_lhs;
    *lhs_ptr='\0';
    buffer_rhs=calloc(sizeof(char),strlen(params));
    rhs_ptr=buffer_rhs;
    *rhs_ptr='\0';

    for(ptr=params;*ptr!='\0';ptr++){
      switch(*ptr){
      case '=':
	// reset buffer
	rhs_ptr=buffer_rhs;
	*rhs_ptr='\0';
	lhs=false;
	break;
      case ';':
	//set parameter
	ret=set_parameter(buffer_lhs,buffer_rhs,K1,K2,N1,N2,nsteps,nu,delta,L,print_freq,&setN1,&setN2);
	if(ret<0){
	  return ret;
	}
	// reset buffer
	lhs_ptr=buffer_lhs;
	*lhs_ptr='\0';
	lhs=true;
	break;
      default:
	// add to buffer
	if (lhs){
	  *lhs_ptr=*ptr;
	  lhs_ptr++;
	  *lhs_ptr='\0';
	}
	else{
	  *rhs_ptr=*ptr;
	  rhs_ptr++;
	  *rhs_ptr='\0';
	}
	break;
      }
    }

    // set last param
    if (*params!='\0'){
      ret=set_parameter(buffer_lhs,buffer_rhs,K1,K2,N1,N2,nsteps,nu,delta,L,print_freq,&setN1,&setN2);
      if(ret<0){
	return ret;
      }
    }

    // free vects
    free(buffer_lhs);
    free(buffer_rhs);
  }

  // if N not set explicitly, set it
  if (!setN1){
    *N1=4*(*K1)+1;
  }
  if (!setN2){
    *N2=4*(*K2)+1;
  }

  return(0);
}


// set a parameter from the parameter string
int set_parameter(
  char* lhs,
  char* rhs,
  int* K1,
  int* K2,
  int* N1,
  int* N2,
  unsigned int* nsteps,
  double* nu,
  double* delta,
  double* L,
  unsigned int* print_freq,
  bool* setN1,
  bool* setN2
){
  int ret;

  if (strcmp(lhs,"K1")==0){
    ret=sscanf(rhs,"%d",K1);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'K1' should be an integer\n       got '%s'\n",rhs);
      return(-1);
    }
  }
  else if (strcmp(lhs,"K2")==0){
    ret=sscanf(rhs,"%d",K2);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'K2' should be an integer\n       got '%s'\n",rhs);
      return(-1);
    }
  }
  else if (strcmp(lhs,"K")==0){
    ret=sscanf(rhs,"%d",K1);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'K' should be an integer\n       got '%s'\n",rhs);
      return(-1);
    }
    *K2=*K1;
  }
  else if (strcmp(lhs,"N1")==0){
    ret=sscanf(rhs,"%d",N1);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'N1' should be an integer\n       got '%s'\n",rhs);
      return(-1);
    }
    *setN1=true;
  }
  else if (strcmp(lhs,"N2")==0){
    ret=sscanf(rhs,"%d",N2);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'N2' should be an integer\n       got '%s'\n",rhs);
      return(-1);
    }
    *setN2=true;
  }
  else if (strcmp(lhs,"N")==0){
    ret=sscanf(rhs,"%d",N1);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'N' should be an integer\n       got '%s'\n",rhs);
      return(-1);
    }
    *N2=*N1;
    *setN1=true;
    *setN2=true;
  }
  else if (strcmp(lhs,"nsteps")==0){
    ret=sscanf(rhs,"%u",nsteps);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'nsteps' should be an unsigned integer\n       got '%s'\n",rhs);
      return(-1);
    }
  }
  else if (strcmp(lhs,"nu")==0){
    ret=sscanf(rhs,"%lf",nu);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'nu' should be a double\n       got '%s'\n",rhs);
      return(-1);
    }
  }
  else if (strcmp(lhs,"delta")==0){
    ret=sscanf(rhs,"%lf",delta);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'delta' should be a double\n       got '%s'\n",rhs);
      return(-1);
    }
  }
  else if (strcmp(lhs,"L")==0){
    ret=sscanf(rhs,"%lf",L);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'L' should be a double\n       got '%s'\n",rhs);
      return(-1);
    }
  }
  else if (strcmp(lhs,"print_freq")==0){
    ret=sscanf(rhs,"%u",print_freq);
    if(ret!=1){
      fprintf(stderr, "error: parameter 'print_freq' should be an unsigned integer\n       got '%s'\n",rhs);
      return(-1);
    }
  }
  else{
    fprintf(stderr, "error: unrecognized parameter '%s'\n",lhs);
    return(-1);
  }

  return(0);
}
